/*
 This file is part of the digitalSTROM ds485-client library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-client library is free software: you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 digitalSTROM ds485-client library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public License
 along with digitalSTROM ds485-client library. If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <assert.h>
#include <syslog.h>

#include <digitalSTROM/ds485-socket-server.h>
#include <digitalSTROM/ds485-socket-interface.h>
#include "ds485-client-interface.h"
#include "ds485-client.h"

#include "base/log.h"

static int ds485_client_connect(ds485ClientHandle_t handle,
                                const char *connspec, int flags);
static int ds485_parse_connection_url(ds485_client_t *cl, const char *connspec);

ds485ClientHandle_t ds485_client_create()
{
    ds485_client_t *cl;

    cl = calloc(1, sizeof(ds485_client_t));
    if (cl == NULL) {
        return NULL;
    }

    cl->busState = DS485_STATE_ISOLATED;
    cl->shutdown = false;

    waitqueue_create(&cl->syncSendShort.wq, "sync_short");
    waitqueue_create(&cl->syncSendLong.wq, "sync_long");
    waitqueue_create(&cl->busQuery.wq, "bus_query");
    waitqueue_create(&cl->openQuery.wq, "open");
    return cl;
}

/**
 * Will be executed automatically when thread exits
 *
 * search for pthread_cleanup_pop()
 */
static void ds485_cleanup(ds485_client_t *cl)
{
    waitqueue_destroy(&cl->syncSendShort.wq);
    waitqueue_destroy(&cl->syncSendLong.wq);
    waitqueue_destroy(&cl->busQuery.wq);
    waitqueue_destroy(&cl->openQuery.wq);

    free(cl->dName);
    free(cl->dPort);
    free(cl);
}

/**
 * Opens a connection using the specified connection type
 *
 * @CONNSPEC: tcp://localhost:8442
 * @flags: PROMISCUOUS_MODE - server shall forward all messages
 * @RETURN: handle or NULL on failure
 *
 * @deprecated: use ds485_client_open2
 */
PUBLIC_API ds485ClientHandle_t ds485_client_open(const char *connspec, int flags)
{
    return ds485_client_open2(connspec, flags, NULL);
}

/**
 * Creates a handle, installs callbacks and connects to server
 *
 * @CONNSPEC: tcp://localhost:8442
 * @flags: PROMISCUOUS_MODE - server shall forward all messages
 * @callbacks: state / bus_change / long and short messages
 *
 * @RETURN: handle or NULL on failure
 */
PUBLIC_API ds485ClientHandle_t ds485_client_open2(const char *connspec, int flags,
                                       struct ds485c_callbacks *callbacks)
{
    ds485_client_t *cl;
    int rval;

    cl = ds485_client_create();
    if (cl == NULL) {
        return NULL;
    }

    if (callbacks) {
        memcpy(&cl->callbacks, callbacks, sizeof(*callbacks));
    }

    rval = ds485_client_connect(cl, connspec, flags);
    if (rval < 0) {
        (void)ds485_client_close(cl);
        return NULL;
    }

    return cl;
}

static int ds485_parse_connection_url(ds485_client_t *cl, const char *connspec)
{
    const char *name, *port;
    char *tmp;

    if (strncmp(connspec, "tcp://", 6)) {
        return ERROR_INVALID_CONNSPEC;
    }

    /* strtok modifies first argument */
    tmp = strdup(connspec + 6);
    name = strtok(tmp, ":");
    if (name == NULL) {
        return ERROR_INVALID_CONNSPEC;
    }
    port = strtok(NULL, ":");
    if (port == NULL) {
        return ERROR_INVALID_CONNSPEC;
    }
    cl->dProtocol = "tcp";
    cl->dName = strdup(name);
    cl->dPort = strdup(port);
    free(tmp);

    update_pretty_name(cl);
    return 0;
}

static int ds485_client_connect(ds485ClientHandle_t handle,
                                const char *connspec, int flags)
{
    ds485_client_t *cl = handle;
    int rval;

    cl->sockfd = -1;
    cl->flags = flags;

    rval = ds485_parse_connection_url(cl, connspec);
    if (rval < 0) {
        return rval;
    }

    waitqueue_enqueue(&cl->openQuery.wq);
    rval = ds485_socket_client_open(cl);
    if (rval < 0) {
        waitqueue_dequeue(&cl->openQuery.wq);
        return rval;
    }

    rval = waitqueue_wait(&cl->openQuery.wq, 10);
    waitqueue_dequeue(&cl->openQuery.wq);
    return rval;
}

/**
 * Closes the session, frees all connection related resources
 */
PUBLIC_API int ds485_client_close(ds485ClientHandle_t handle)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    int rval;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    cl->shutdown = true;
    rval = pthread_join(cl->connThread, NULL);
    if (rval) {
        /* thread probably wasn't started... nevermind */
    }
    ds485_cleanup(cl);
    return 0;
}

/**
 * Query the list of devices on the dS485 bus
 */
PUBLIC_API int ds485_client_query_devices(ds485ClientHandle_t handle, dsuid_t *device_list, int device_list_size)
{
    ds485_client_t *cl = handle;
    int rval;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    waitqueue_enqueue(&cl->busQuery.wq);

    cl->busQuery.type = query_devices;
    cl->busQuery.devices = device_list;
    cl->busQuery.length = device_list_size;

    rval = ds485_socket_client_get_bus_devices(cl, device_list, device_list_size);
    if (rval < 0) {
        dlog(LOG_NOTICE, cl, "get_bus_devices: %s\n", ds485c_strerror(rval));
        waitqueue_dequeue(&cl->busQuery.wq);
        return rval;
    }

    rval = waitqueue_wait(&cl->busQuery.wq, 10);
    if (rval < 0) {
        waitqueue_dequeue(&cl->busQuery.wq);
        return rval;
    }

    rval = cl->busQuery.length;
    waitqueue_dequeue(&cl->busQuery.wq);
    return rval;
}

/**
 * Query the current bus state
 */
PUBLIC_API int ds485_client_query_state(ds485ClientHandle_t handle, bus_state_t *state)
{
    ds485_client_t *cl = handle;
    int rval;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    waitqueue_enqueue(&cl->busQuery.wq);

    cl->busQuery.type = query_state;
    rval = ds485_socket_client_get_bus_state(cl, state);
    if (rval < 0) {
        dlog(LOG_NOTICE, cl, "bus_state: %s\n", ds485c_strerror(rval));
        waitqueue_dequeue(&cl->busQuery.wq);
        return rval;
    }

    rval = waitqueue_wait(&cl->busQuery.wq, 2);
    if (rval < 0) {
        waitqueue_dequeue(&cl->busQuery.wq);
        *state = DS485_STATE_DISCONNECTED;
        return rval;
    }

    *state = cl->busQuery.length;
    waitqueue_dequeue(&cl->busQuery.wq);
    return 0;
}

PUBLIC_API int ds485_client_restart_bus(ds485ClientHandle_t handle)
{
    ds485_client_t *cl = handle;
    int rval;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    waitqueue_enqueue(&cl->busQuery.wq);

    cl->busQuery.type = restart_bus;
    rval = ds485_socket_client_restart_bus(cl);
    if (rval < 0) {
        dlog(LOG_NOTICE, cl, "restart_bus: %s\n", ds485c_strerror(rval));
        waitqueue_dequeue(&cl->busQuery.wq);
        return rval;
    }

    waitqueue_wait(&cl->busQuery.wq, 2);
    if (rval < 0) {
        waitqueue_dequeue(&cl->busQuery.wq);
        return rval;
    }

    waitqueue_dequeue(&cl->busQuery.wq);
    return 0;
}

/**
 * Send dM-API command
 */
PUBLIC_API int ds485_client_send_command(ds485ClientHandle_t handle,
                              const ds485_container_t *request)
{
    ds485_client_t *cl = (ds485_client_t *)handle;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    if (request->length > DS485_MAX_PAYLOAD_SIZE) {
        dlog(LOG_ERR, cl, "refusing to send oversize container: %d\n",
             request->length);
        return ERROR_SOCKET_FULL;
    }

    return ds485_socket_client_send_command(cl, request);
}

PUBLIC_API int ds485_client_send_long_command(ds485ClientHandle_t handle,
                                   const ds485n_packet_t *request)
{
    ds485_client_t *cl = (ds485_client_t *)handle;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    if (request->length > DS485N_PACKET_MAX_PAYLOAD_SIZE) {
        dlog(LOG_ERR, cl, "refusing to send oversize long container: %d\n",
             request->length);
        return ERROR_SOCKET_FULL;
    }

    return ds485_socket_client_send_long_command(cl, request);
}

/**
 * Send dM-API command, receive a single response packet
 */
PUBLIC_API int ds485_client_send_sync_command(ds485ClientHandle_t handle, ds485_container_t *request, ds485_container_t *response, int timeout)
{
    ds485_client_t *cl = handle;
    ds485_container_t *tmp;
    int rval;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    waitqueue_enqueue(&cl->syncSendShort.wq);

    response->sourceId = request->destinationId;
    response->containerType = DS485_CONTAINER_RESPONSE;
    response->data[0] = request->data[0];
    response->data[1] = request->data[1];
    cl->syncSendShort.response = *response;

    rval = ds485_socket_client_send_command(cl, request);
    if (rval < 0) {
        dlog(LOG_NOTICE, cl, "send_command: %s\n", ds485c_strerror(rval));
        tmp = &cl->syncSendShort.response;
        memset(tmp, 0, sizeof(*tmp));
        waitqueue_dequeue(&cl->syncSendShort.wq);
        return rval;
    }

    if (dsuid_is_broadcast(&request->destinationId)) {
        tmp = &cl->syncSendShort.response;
        memset(tmp, 0, sizeof(*tmp));
        waitqueue_dequeue(&cl->syncSendShort.wq);
        return NOTICE_CALL_TO_BROADCAST;
    }

    rval = waitqueue_wait(&cl->syncSendShort.wq, timeout);
    if (rval < 0) {
        tmp = &cl->syncSendShort.response;
        memset(tmp, 0, sizeof(*tmp));
        waitqueue_dequeue(&cl->syncSendShort.wq);
        return rval;
    }

    *response = cl->syncSendShort.response;
    tmp =  &cl->syncSendShort.response;
    memset(tmp, 0, sizeof(*tmp));
    waitqueue_dequeue(&cl->syncSendShort.wq);
    return rval;
}

PUBLIC_API int ds485_client_send_sync_long_command(ds485ClientHandle_t handle, ds485n_packet_t *request, ds485n_packet_t *response, int timeout)
{
    ds485_client_t *cl = handle;
    ds485n_packet_t *tmp;
    int rval;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    waitqueue_enqueue(&cl->syncSendLong.wq);

    response->sourceId = request->destinationId;
    response->containerType = DS485_CONTAINER_RESPONSE;
    response->data[0] = request->data[0];
    response->data[1] = request->data[1];
    cl->syncSendLong.response = *response;

    rval = ds485_socket_client_send_long_command(cl, request);
    if (rval < 0) {
        dlog(LOG_NOTICE, cl, "send_long_command: %s\n", ds485c_strerror(rval));
        tmp =  &cl->syncSendLong.response;
        memset(tmp, 0, sizeof(*tmp));
        waitqueue_dequeue(&cl->syncSendLong.wq);
        return rval;
    }

    if (dsuid_is_broadcast(&request->destinationId)) {
        tmp =  &cl->syncSendLong.response;
        memset(tmp, 0, sizeof(*tmp));
        waitqueue_dequeue(&cl->syncSendLong.wq);
        return NOTICE_CALL_TO_BROADCAST;
    }

    rval = waitqueue_wait(&cl->syncSendLong.wq, timeout);
    if (rval < 0) {
        tmp =  &cl->syncSendLong.response;
        memset(tmp, 0, sizeof(*tmp));
        waitqueue_dequeue(&cl->syncSendLong.wq);
        return rval;
    }

    *response = cl->syncSendLong.response;
    tmp =  &cl->syncSendLong.response;
    memset(tmp, 0, sizeof(*tmp));
    waitqueue_dequeue(&cl->syncSendLong.wq);
    return 0;
}

/**
 * Set and remove a reception handler for dS485 Packet's
 */
PUBLIC_API client_callback_container_t ds485_client_set_receive_callback(ds485ClientHandle_t handle, client_callback_container_t cb, void *data)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    if (cl == NULL) {
        return NULL;
    }
    client_callback_container_t prev_cb = cl->callbacks.container_pkt_cb;
    cl->callbacks.container_pkt_cb = cb;
    cl->callbacks.container_pkt_data = data;
    return prev_cb;
}

PUBLIC_API client_callback_netlib_packet_t ds485_client_set_receive_netlib_packet_callback(ds485ClientHandle_t handle, client_callback_netlib_packet_t cb, void *data)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    if (cl == NULL) {
        return NULL;
    }
    client_callback_netlib_packet_t prev_cb = cl->callbacks.netlib_pkt_cb;
    cl->callbacks.netlib_pkt_cb = cb;
    cl->callbacks.netlib_pkt_data = data;
    return prev_cb;
}

/**
 * Set and remove a callback handler to register for bus state changes
 */
PUBLIC_API client_callback_state_t ds485_client_set_state_callback(ds485ClientHandle_t handle, client_callback_state_t cb, void *data)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    if (cl == NULL) {
        return NULL;
    }
    client_callback_state_t prev_cb = cl->callbacks.state_change_cb;
    cl->callbacks.state_change_cb = cb;
    cl->callbacks.state_change_data = data;
    return prev_cb;
}

PUBLIC_API client_callback_bus_change_t ds485_client_set_bus_change_callback(ds485ClientHandle_t handle, client_callback_bus_change_t cb, void *data)
{
    ds485_client_t *cl = (ds485_client_t *)handle;
    if (cl == NULL) {
        return NULL;
    }
    client_callback_bus_change_t prev_cb = cl->callbacks.bus_change_cb;
    cl->callbacks.bus_change_cb = cb;
    cl->callbacks.bus_change_data = data;
    return prev_cb;
}

PUBLIC_API int ds485_client_get_dsuid(ds485ClientHandle_t handle, dsuid_t *dsuid)
{
    ds485_client_t *cl = handle;
    dsuid_t local_dsuid;
    int rval;

    if (cl == NULL) {
        return ERROR_INVALID_HANDLE;
    }

    waitqueue_enqueue(&cl->busQuery.wq);

    cl->busQuery.type = query_dsuid;
    cl->busQuery.devices = &local_dsuid;

    rval = ds485_socket_client_get_dsuid(cl);
    if (rval < 0) {
        dlog(LOG_NOTICE, cl, "get_dsuid: %s\n", ds485c_strerror(rval));
        waitqueue_dequeue(&cl->busQuery.wq);
        return rval;
    }

    rval = waitqueue_wait(&cl->busQuery.wq, 2);
    if (rval < 0) {
        waitqueue_dequeue(&cl->busQuery.wq);
        *dsuid = DSUID_NULL;
        return rval;
    }

    *dsuid = cl->busQuery.devices[0];
    waitqueue_dequeue(&cl->busQuery.wq);
    return rval;
}

static void init(void) __attribute__((constructor));
static void init(void) {
    const char *env;
    int loglevel;

    init_logger("ds485c", LOG_NOTICE, false);

    env = getenv(DS485_DEBUG);
    if (env) {
        errno = 0;
        loglevel = strtol(env, NULL, 10);
        if (errno) {
            report(LOG_WARNING, "Invalid DS485_DEBUG=%s, ingoring\n",
                   env);
        } else {
            set_log_level(loglevel);
        }
    }

    env = getenv(DS485_SYSLOG);
    if (env) {
        enable_syslog(true);
    }
}

static void fini(void) __attribute__((destructor));
static void fini(void) {
}
