/*
 * Copyright (C) 2014 digitalSTROM.org, Schlieren, Switzerland
 *
 * digitalSTROM ds485-client library is free software: you can redistribute it
 * and/or modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 *
 * digitalSTROM ds485-client library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with digitalSTROM ds485-client library. If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "ds485-client.h"

#include <assert.h>
#include <errno.h>
#include <pthread.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/time.h>

#include "base/log.h"

void waitqueue_create(waitqueue_t *wq, const char *name)
{
    pthread_mutex_init(&wq->mutex, NULL);
    pthread_cond_init(&wq->condition, NULL);
    wq->state = STATE_IDLE;
    wq->name = strdup(name);
}

void waitqueue_destroy(waitqueue_t *wq)
{
    pthread_mutex_destroy(&wq->mutex);
    pthread_cond_destroy(&wq->condition);
    free(wq->name);
}

int waitqueue_enqueue(waitqueue_t *wq)
{
    pthread_mutex_lock(&wq->mutex);
    assert(wq->state == STATE_IDLE);
    wq->state = STATE_WAITING;
    return 0;
}

int waitqueue_wait(waitqueue_t *wq, int timeout_sec)
{
    struct timespec ts;
    struct timeval now;
    int rval = 0;

    /* TODO clock_gettime introduces -lrt dependency */
    gettimeofday(&now, NULL);
    ts.tv_sec = now.tv_sec + timeout_sec;
    ts.tv_nsec = now.tv_usec * 1000;

    while ((rval == 0) && (wq->state != STATE_RECEIVED)) {
        rval = pthread_cond_timedwait(&wq->condition, &wq->mutex, &ts);
        if (rval) {
            report(LOG_WARNING, "wq:%s pthread_cond_timedwait:%s(%d) state:%d\n",
                   wq->name, strerror(rval), rval, wq->state);
        } else if (wq->state != STATE_RECEIVED){
            report(LOG_INFO, "wq:%s pthread_cond_timedwait: sporadic wakeup\n",
                   wq->name);
        }
    }

    if (wq->state == STATE_RECEIVED) {
        return 0;
    } else if (rval != ETIMEDOUT) {
        return ERROR_WAITQ_USAGE;
    } else {
        return ERROR_RESPONSE_TIMEOUT;
    }
}

int waitqueue_dequeue(waitqueue_t *wq)
{
    wq->state = STATE_IDLE;
    pthread_mutex_unlock(&wq->mutex);
    return 0;
}

/**
 * waitqueue_waiting
 * @return true if thread is waiting for signal
 * must pair with waitqueue_unlock
 */
bool waitqueue_waiting(waitqueue_t *wq)
{
    pthread_mutex_lock(&wq->mutex);
    return (wq->state == STATE_WAITING);
}

int waitqueue_signal(waitqueue_t *wq)
{
    wq->state = STATE_RECEIVED;
    pthread_cond_signal(&wq->condition);
    return 0;
}

int waitqueue_unlock(waitqueue_t *wq)
{
    pthread_mutex_unlock(&wq->mutex);
    return 0;
}
